#!/usr/bin/env ruby

require 'gtk2'
require 'webkit'
require 'fileutils'
require 'rbconfig'

class Editor < Gtk::Window
  attr_accessor :filename

  def exit
    @list.delete(self)
    self.destroy
  end

  def initialize(title, filename, list, tree, window)
    super(title)
    resize(700, 500)
    @filename = filename
    @tree = tree
    @window = window
    @list = list

    @on_exit = Proc.new do
      if File.exists?(@filename) and get_html != File.read(@filename)
        dialog = Gtk::MessageDialog.new(nil,
                                        Gtk::Dialog::DESTROY_WITH_PARENT,
                                        Gtk::MessageDialog::WARNING,
                                        Gtk::MessageDialog::BUTTONS_NONE,
                                        'Document was changed')
        dialog.add_buttons([Gtk::Stock::SAVE, 0],
                           [Gtk::Stock::DISCARD, 1],
                           [Gtk::Stock::CANCEL, 2])
        dialog.secondary_text = 'Do you want to save changes?'
        dialog.signal_connect('response') { |dialog, response|
          case response
          when 0 # save
            @on_save.call
            exit
          when 1 # discard
            exit
          end
          dialog.destroy
        }
        dialog.show_all
      else
        @on_save.call
        exit
      end
    end

    signal_connect("delete_event", &@on_exit)

    @editor = Gtk::WebKit::WebView.new
    @editor.editable = true
    if File.exists?(@filename)
      @editor.open(File.expand_path(@filename))
#      File.open(@filename) { |fd|
#        @editor.load_html_string(fd.read(), '')
#      }
    else
      @editor.load_html_string('', '')
    end

    scroll = Gtk::ScrolledWindow.new
    scroll.add(@editor)
    scroll.set_policy(Gtk::POLICY_AUTOMATIC, Gtk::POLICY_AUTOMATIC)

    color_set_cb = Proc.new do |colorbutton|
      gc = colorbutton.color.to_s
      color = "##{gc[1...3]}#{gc[5...7]}#{gc[9...11]}"
      @editor.execute_script("document.execCommand('%s', null, '%s');" % [colorbutton.name, color])
    end
    @colorbutton = Gtk::ColorButton.new
    @colorbutton.name = "forecolor"
    @colorbutton.title = 'Select foreground color'
    @colorbutton.signal_connect('color-set', &color_set_cb)

    @colorbgbutton = Gtk::ColorButton.new
    @colorbgbutton.name = "backcolor"
    @colorbgbutton.title = 'Select background color'
    @colorbgbutton.signal_connect('color-set', &color_set_cb)

    @ui = generate_ui
    add_accel_group(@ui.accel_group)
    @toolbar1 = @ui.get_widget("/toolbar_main")
    @toolbar1.toolbar_style = Gtk::Toolbar::Style::ICONS
    @toolbar2 = @ui.get_widget("/toolbar_format")
    @toolbar2.toolbar_style = Gtk::Toolbar::Style::ICONS
    @toolbar3 = @ui.get_widget("/toolbar_attach")
    @toolbar3.toolbar_style = Gtk::Toolbar::Style::ICONS
    @menubar = @ui.get_widget("/menubar_main")

    store = Gtk::ListStore.new(String)
    @attach = Gtk::TreeView.new(store)
    @attach.insert_column(0, 'Attachments', Gtk::CellRendererText.new,
                          { :text => 0 })
    attachmentsscr = Gtk::ScrolledWindow.new
    attachmentsscr.set_size_request(150, -1)
    attachmentsscr.set_policy('automatic', 'automatic')
    attachmentsscr.add(@attach)
    refresh

    @attach.signal_connect('row-activated') {
      iter = @attach.selection.selected
      open_file(@tree, iter[0])
    }

    popup = proc { |path, event|
      menu = Gtk::Menu.new

      item = Gtk::ImageMenuItem.new(Gtk::Stock::ADD)
      item.signal_connect('activate', &@on_attach)
      menu.append(item)

      if path
        iter = @attach.model.get_iter(path)

        item = Gtk::ImageMenuItem.new(Gtk::Stock::OPEN)
        item.signal_connect('activate') {
          open_file(@tree, iter[0])
        }
        menu.append(item)

        item = Gtk::ImageMenuItem.new(Gtk::Stock::SAVE)
        item.signal_connect('activate') {
          iter = @attach.selection.selected

          dialog = Gtk::FileChooserDialog.new("Save attachment", self,
                                              Gtk::FileChooser::ACTION_SAVE, nil,
                                              [Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_CANCEL],
                                              [Gtk::Stock::SAVE, Gtk::Dialog::RESPONSE_OK])
          dialog.current_name = iter[0]

          if dialog.run == Gtk::Dialog::RESPONSE_OK
            fn = dialog.filename
            FileUtils.cp([@tree.path, iter[0]].join('/'), fn)
          end
          dialog.destroy()
        }
        menu.append(item)

        item = Gtk::ImageMenuItem.new(Gtk::Stock::DELETE)
        item.signal_connect('activate', &@on_delete)
        menu.append(item)
      end

      menu.show_all

      if event              # mouse
        @attach.set_cursor(path, nil, false) if path
        menu.popup(nil, nil, event.button, event.time)
      else                  # keyboard
        menu.popup(nil, nil, 0, Gdk::Event::CURRENT_TIME)
      end
    }
    @attach.signal_connect("popup_menu") { |widget|
      path, *tmp = @attach.cursor
      popup.call(path, nil)
    }
    @attach.signal_connect('button_press_event') { |widget, event|
      if event.kind_of? Gdk::EventButton and event.button == 3
        path, *tmp = @attach.get_path(event.x, event.y)
        popup.call(path, event)
      end
    }


    hpan = Gtk::HPaned.new
    v = Gtk::VBox.new
    v.pack_start(@toolbar2, false)
    v.pack_start(scroll)

    hpan.pack1(v, true, false)

    v = Gtk::VBox.new
    v.pack_start(@toolbar3, false)
    v.pack_start(attachmentsscr)

    hpan.pack2(v, false, true)


    @layout = Gtk::VBox.new
    @layout.pack_start(@menubar, false)
    @layout.pack_start(@toolbar1, false)
    # @layout.pack_start(@toolbar2, false)
    @layout.pack_start(hpan, true, true)

    add(@layout)
    show_all
    @editor.grab_focus
  end

  def refresh
    @attach.model.clear

    files = []
    Dir["#{@tree.path}/*"].each do |pathname|
      basename = File.basename(pathname)
      if not FileTest.directory?(pathname) and basename != NOTE
        files << basename
      end
    end

    files.sort!.each { |fn| @attach.model.append[0] = fn }
  end

  def generate_ui
    ui_def = <<END
    <ui>
      <menubar name="menubar_main">
        <menu action="menuFile">
          <menuitem action="save" />
          <menuitem action="print" />
          <separator />
          <menuitem action="attach" />
        </menu>
        <menu action="menuEdit">
          <menuitem action="undo" />
          <menuitem action="redo" />
          <separator />
          <menuitem action="cut" />
          <menuitem action="copy" />
          <menuitem action="paste" />
          <separator />
          <menuitem action="search" />
        </menu>
<!--
        <menu action="menuInsert">
          <menuitem action="insertimage" />
        </menu>
-->
        <menu action="menuFormat">
          <menuitem action="bold" />
          <menuitem action="italic" />
          <menuitem action="underline" />
          <menuitem action="strikethrough" />
          <separator />
          <menuitem action="font" />
          <menuitem action="color" />
          <menuitem action="colorbg" />
          <separator />
          <menuitem action="justifyleft" />
          <menuitem action="justifyright" />
          <menuitem action="justifycenter" />
          <menuitem action="justifyfull" />
        </menu>
      </menubar>
      <toolbar name="toolbar_main">
        <!--
        <toolitem action="new" />
        <toolitem action="open" />
        -->
        <toolitem action="save" />
        <separator />
        <toolitem action="undo" />
        <toolitem action="redo" />
        <separator />
        <toolitem action="cut" />
        <toolitem action="copy" />
        <toolitem action="paste" />
      </toolbar>
      <toolbar name="toolbar_format">
        <toolitem action="bold" />
        <toolitem action="italic" />
        <toolitem action="underline" />
        <toolitem action="strikethrough" />
        <separator />
        <toolitem action="font" />
        <placeholder name="size" action="size" />
        <toolitem action="color" />
        <toolitem action="colorbg" />
        <separator />
        <toolitem action="justifyleft" />
        <toolitem action="justifycenter" />
        <toolitem action="justifyright" />
        <toolitem action="justifyfull" />
<!--
        <separator />
        <toolitem action="insertimage" />
-->
      </toolbar>

      <toolbar name="toolbar_attach">
        <toolitem action="attach" />
      </toolbar>
    </ui>
END

    on_action = Proc.new do |actiongroup, action|
      @editor.execute_script("document.execCommand('%s', false, false);" % action.name)
    end

    on_paste = Proc.new do |actiongroup, action|
      @editor.paste_clipboard
    end

    #on_new = Proc.new do |actiongroup, action|
    #  @editor.load_html_string("", "file:///")
    #end

    #on_insert_image = Proc.new do |actiongroup, action|
    #  dialog = Gtk::FileChooserDialog.new("Select an image file", self, Gtk::FileChooser::ACTION_OPEN, nil,
    #                                  [Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_CANCEL], [Gtk::Stock::OPEN, Gtk::Dialog::RESPONSE_OK])
    #
    #  if dialog.run == Gtk::Dialog::RESPONSE_OK
    #    fn = dialog.filename
    #    if File.exists?(fn)
    #      @editor.execute_script("document.execCommand('insertImage', null, '%s');" % fn)
    #    end
    #  end
    #  dialog.destroy()
    #end

    #on_open = Proc.new do |actiongroup, action|
    #  dialog = Gtk::FileChooserDialog.new("Select an HTML file", self, Gtk::FileChooser::ACTION_OPEN, nil,
    #                                      [Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_CANCEL],
    #                                      [Gtk::Stock::OPEN, Gtk::Dialog::RESPONSE_OK])
    #
    #  if dialog.run() == Gtk::Dialog::RESPONSE_OK
    #    fn = dialog.filename
    #    if File.exists?(fn)
    #      @filename = fn
    #      File.open(fn) { |fd|
    #        @editor.load_html_string(fd.read(), "file:///")
    #      }
    #    end
    #  end
    #  dialog.destroy()
    #end

    @on_save = Proc.new do |actiongroup, action|
      if @filename
        File.open(@filename, "w+") { |fd|
          fd.write(get_html)
        }
        @tree.parent.store.each { |model, path, iter|
          if iter[TREE] == @tree
            iter[ICON] = Gtk::Stock::YES
            break
          end
        }
        @window.path.on_save(@filename.split('/')[0...-1].join('/'))
#      else
#        dialog = Gtk::FileChooserDialog.new("Select an HTML file", self,
#                                        Gtk::FileChooser::ACTION_SAVE, nil,
#                                        [Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_CANCEL],
#                                            [Gtk::Stock::SAVE, Gtk::Dialog::RESPONSE_OK])
#
#        if dialog.run() == Gtk::Dialog::RESPONSE_OK
#          @filename = dialog.filename()
#          File.open(@filename, "w+") { |fd|
#            fd.write(get_html)
#          }
#        end
#        dialog.destroy()
      end
    end

    on_select_font = Proc.new do |actiongroup, action|
      FontDialog.new(self) { |font, size|
        if font
          @editor.execute_script("document.execCommand('fontname', null, '%s');"% font[0].name)
        end
        if size
          @editor.execute_script("document.execCommand('fontsize', null, '%s');"% size[0])
        end
      }
    end
    on_select_color = Proc.new do |actiongroup, action|
      @colorbutton.clicked
    end
    on_select_colorbg = Proc.new do |actiongroup, action|
      @colorbgbutton.clicked
    end

    on_search = Proc.new do |actiongroup, action|
      h = Gtk::HBox.new
      label = Gtk::Label.new('Search phrase:')
      nlabel = Gtk::Label.new
      caseb = Gtk::CheckButton.new('Case sensitive')
      input = Gtk::Entry.new
      h.pack_start(label, false, false)
      h.pack_start(input)

      dialog = Gtk::Dialog.new("Search",
                               self,
                               Gtk::Dialog::DESTROY_WITH_PARENT,
                               [ Gtk::Stock::FIND, 0 ],
                               [ Gtk::Stock::CLOSE, 1])
      dialog.signal_connect('response') { |dialog, response|
        if response == 0
          @editor.unmark_text_matches
          n = @editor.mark_text_matches(input.text, caseb.active?, 0)
          nlabel.text = "%d occurences found." % n
          @editor.highlight_text_matches = true
          @editor.search_text(input.text, caseb.active?, true, true)
        else
          @editor.unmark_text_matches
          dialog.destroy
        end
      }

      dialog.vbox.pack_start(h)
      dialog.vbox.pack_start(nlabel)
      dialog.vbox.pack_start(caseb, false, false)
      dialog.show_all
    end

    on_print = Proc.new do
      @editor.get_main_frame.print
    end

    @on_attach = Proc.new do
      dialog = Gtk::FileChooserDialog.new("Select file", self,
                                          Gtk::FileChooser::ACTION_OPEN, nil,
                                          [Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_CANCEL],
                                          [Gtk::Stock::ADD, Gtk::Dialog::RESPONSE_OK])
      dialog.select_multiple = true

      if dialog.run == Gtk::Dialog::RESPONSE_OK
        fns = dialog.filenames
        fns.each { |fn|
          if File.exists?(fn)
            FileUtils.cp(fn, @tree.path)
          end
        }
        refresh
      end
      dialog.destroy()
    end

    @on_delete = Proc.new do |fn|
      iter = @attach.selection.selected
      dialog = Gtk::MessageDialog.new(self,
                                      Gtk::Dialog::DESTROY_WITH_PARENT,
                                      Gtk::MessageDialog::QUESTION,
                                      Gtk::MessageDialog::BUTTONS_YES_NO,
                                      'Delete')
      dialog.secondary_text = "Are you sure you want to delete \"%s\"?" % iter[0]
      dialog.signal_connect('response') { |dialog, response|
        if response == Gtk::Dialog::RESPONSE_YES
          FileUtils.rm([@tree.path, iter[0]].join('/'))
          @attach.model.remove(iter)
        end
        dialog.destroy
      }
      dialog.show_all
    end

    actions = Gtk::ActionGroup.new("Actions")
    actions.add_actions([
      ["menuFile", nil, "_File"],
      ["menuEdit", nil, "_Edit"],
      #["menuInsert", nil, "_Insert"],
      ["menuFormat", nil, "F_ormat"],

      #["new", Gtk::Stock::NEW, "_New", nil, nil, on_new],
      #["open", Gtk::Stock::OPEN, "_Open", nil, nil, on_open],
      ["save", Gtk::Stock::SAVE, "_Save", nil, nil, @on_save],

      ["attach", Gtk::Stock::ADD, "_Attach", nil, nil, @on_attach],

      ["undo", Gtk::Stock::UNDO, "_Undo", '<ctrl>Z', nil, on_action],
      ["redo", Gtk::Stock::REDO, "_Redo", '<ctrl>Y', nil, on_action],

      ["cut", Gtk::Stock::CUT, "Cu_t", nil, nil, on_action],
      ["copy", Gtk::Stock::COPY, "_Copy", nil, nil, on_action],
      ["paste", Gtk::Stock::PASTE, "_Paste", nil, nil, on_paste],

      ["search", Gtk::Stock::FIND, "_Search", "<ctrl>F", nil, on_search],
      ["print", Gtk::Stock::PRINT, "P_rint", "<ctrl>P", nil, on_print],

      ["bold", Gtk::Stock::BOLD, "_Bold", "<ctrl>B", nil, on_action],
      ["italic", Gtk::Stock::ITALIC, "_Italic", "<ctrl>I", nil, on_action],
      ["underline", Gtk::Stock::UNDERLINE, "_Underline", "<ctrl>U", nil, on_action],
      ["strikethrough", Gtk::Stock::STRIKETHROUGH, "_Strike", "<ctrl>T", nil, on_action],
      ["font", Gtk::Stock::SELECT_FONT, "_Font", "<ctrl>D", nil, on_select_font],
      ["color", Gtk::Stock::SELECT_COLOR, "Text F_oreground", nil, nil, on_select_color],
      ["colorbg", Gtk::Stock::SELECT_COLOR, "Text B_ackground", nil, nil, on_select_colorbg],

      ["justifyleft", Gtk::Stock::JUSTIFY_LEFT, "Justify _Left", "<alt>L", nil, on_action],
      ["justifycenter", Gtk::Stock::JUSTIFY_CENTER, "Justify _Center", "<alt>C", nil, on_action],
      ["justifyright", Gtk::Stock::JUSTIFY_RIGHT, "Justify _Right", "<alt>R", nil, on_action],
      ["justifyfull", Gtk::Stock::JUSTIFY_FILL, "_Justify Full", "<alt>J", nil, on_action]

      #["insertimage", "insert-image", "Insert _Image", nil, nil, on_insert_image]
    ])

    ui = Gtk::UIManager.new
    ui.insert_action_group(actions, 0)
    ui.add_ui(ui_def)
    return ui
  end

  def get_html
    # or use get_main_frame.get_data_source.get_data (since webkit 1.1.14)
    @editor.execute_script("document.title=document.documentElement.innerHTML;")
    html = @editor.get_main_frame.get_title
    @editor.execute_script("document.title='';")
    return html;
  end

end


class FontDialog < Gtk::Dialog
  def initialize(window, &on_exec)
    scr = families_list_view(window)
    super("Choose font", window,
          Gtk::Dialog::DESTROY_WITH_PARENT,
          [ Gtk::Stock::OK, Gtk::Dialog::RESPONSE_OK ],
          [ Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_CANCEL ])
    signal_connect('response') { |dialog, response|
      font = @view.selection.selected
      size = @sizeview.selection.selected
      if response == Gtk::Dialog::RESPONSE_OK
        on_exec.call(font, size)
      end
      dialog.destroy
    }

    activated = Proc.new do
      self.response(Gtk::Dialog::RESPONSE_OK)
    end
    @view.signal_connect('row-activated', &activated)

    sizes = ['xx-small','x-small','small','medium','large','x-large','xx-large']
    store = Gtk::ListStore.new(Integer, String)
    sizes.each_index { |i|
      iter = store.append
      iter[0] = i+1
      iter[1] = sizes[i]
    }
    @sizeview = Gtk::TreeView.new(store)
    @sizeview.insert_column(0, 'Size', Gtk::CellRendererText.new,
                            { :text => 1 })
    # @sizeview.set_cursor(Gtk::TreePath.new('3'), nil, false)
    @sizeview.signal_connect('row-activated', &activated)

    sizescr = Gtk::ScrolledWindow.new
    sizescr.set_size_request(100, -1)
    sizescr.set_policy('automatic', 'automatic')
    sizescr.add(@sizeview)

    hbox = Gtk::HBox.new
    hbox.border_width = 5
    hbox.pack_start(scr)
    hbox.pack_start(sizescr, false, false)
    vbox.pack_start(hbox)
    set_default_size(400, 400)
    show_all
  end

  def families_list(parent)
    store = Gtk::ListStore.new(Pango::FontFamily, String, String)
    families = []
    parent.pango_context.families.each { |f|
      families << [f.name, f]
    }
    families.sort!
    families.each { |name, family|
      iter = store.append
      iter[0] = family
      iter[1] = ''
    }
    return store
  end

  def resize_families_list(size)
    size_str = (size * 1024).to_s
    @store.each do |model, path, row|
      family = row[0]
      name = family.name #CGI::escape(family.name)
      markup = %Q(<span font_family="%s" size="%s">%s</span>\n<span foreground="#666" size="small">(%s)</span>) % [name, size_str, name, name]
      row[2] = name
      row[1] = markup
    end
  end

  def families_list_view(parent)
    @store = families_list(parent)
    resize_families_list(16)
    @view = Gtk::TreeView.new(@store)
    @view.insert_column(0, 'Font', Gtk::CellRendererText.new, { :markup => 1 })
    @view.search_column = 2
    @view.rules_hint = true
    # @view.headers_visible = false
    scrolled = Gtk::ScrolledWindow.new
    scrolled.set_policy('automatic', 'always')
    scrolled.add(@view)
    return scrolled
  end
end
